%{
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <list.h>
#include <hash_map.h>
//#include <map.h>
#include "defs.hh"
#include "gate.hh"

extern int cplex();
extern void cperror(const char *, ...);
hash_map<char *, Gate *, hash<char*>, eqstr> cp_named_gates;
//map<char *, Gate *, ltstr> cp_named_gates;
list<char*> cp_true_gate_names;
list<char*> cp_false_gate_names;
static void addGate(char *, Gate *);
%}

%union {
  char *charptr;
  int intval;
  Gate *gate;
  list<Gate*> *list;
}

%token LPAREN, RPAREN, LBRACKET, RBRACKET
%token SEMICOLON, COMMA, DEF, ASSIGN
%nonassoc EQUIVf, IMPLYf, ORf, ANDf, ODDf, EVENf, NOTf, ITEf
%left EQUIV
%left IMPLY
%left OR
%left AND
%left ODD
%left NOT
%nonassoc <charptr> ID
%token TRUE, FALSE
%nonassoc <intval> NUM

%type <gate> formula
%type <list> formula_list
%type <gate> treshold_rule

%start circuit
%%

circuit:
	  gate_def
	| circuit gate_def
	;

gate_def:
	  ID SEMICOLON {addGate($1,0); }
	| ID DEF formula SEMICOLON { addGate($1, $3); }
        | ASSIGN set_list SEMICOLON
	;

set_list:
          ID                            {cp_true_gate_names.push_back($1); }
        | NOT ID                        {cp_false_gate_names.push_back($2); }
        | set_list COMMA ID             {cp_true_gate_names.push_back($3); }
        | set_list COMMA NOT ID         {cp_false_gate_names.push_back($4); }
        ;

formula:
	  formula EQUIV formula	{$$ = new Gate(Gate::tEQUIV, $1, $3); }
	| formula IMPLY formula	{$$ = new Gate(Gate::tOR, new Gate(Gate::tNOT, $1), $3); }
	| formula OR formula	{$$ = new Gate(Gate::tOR, $1, $3); }
	| formula AND formula	{$$ = new Gate(Gate::tAND, $1, $3); }
	| formula ODD formula	{$$ = new Gate(Gate::tODD, $1, $3); }
	| NOT formula		{$$ = new Gate(Gate::tNOT, $2); }
	| LPAREN formula RPAREN	{$$ = $2; }
	| ID
	  { Gate *existing_gate = cp_named_gates[$1];
	    if(existing_gate) {
	      $$ = existing_gate; free($1); }
	    else {
	      existing_gate = new Gate(Gate::tUNDEF, $1);
	      cp_named_gates[$1] = existing_gate;
	      $$ = existing_gate;
	    }
	  }
	| TRUE			{$$ = new Gate(Gate::tTRUE); }
	| FALSE			{$$ = new Gate(Gate::tFALSE); }
	| EQUIVf LPAREN formula_list RPAREN
	  {$$ = new Gate(Gate::tEQUIV, $3); }
	| IMPLYf LPAREN formula COMMA formula RPAREN
	  {$$ = new Gate(Gate::tOR, new Gate(Gate::tNOT, $3), $5); }
	| ITEf LPAREN formula COMMA formula COMMA formula RPAREN
	  {$$ = new Gate(Gate::tITE, $3, $5, $7); }
	| ORf LPAREN formula_list RPAREN
	  {$$ = new Gate(Gate::tOR, $3); }
	| ANDf LPAREN formula_list RPAREN
	  {$$ = new Gate(Gate::tAND, $3); }
	| EVENf LPAREN formula_list RPAREN
	  {$$ = new Gate(Gate::tEVEN, $3); }
	| ODDf LPAREN formula_list RPAREN
	  {$$ = new Gate(Gate::tODD, $3); }
	| NOTf LPAREN formula RPAREN
	  {$$ = new Gate(Gate::tNOT, $3); }
	| treshold_rule
	  {$$ = $1; }
        ;

formula_list:
	  formula	{$$ = new list<Gate*>(); $$->push_back($1); }
	| formula_list COMMA formula	{$$ = $1; $$->push_back($3); }
	;

treshold_rule:
	LBRACKET NUM COMMA NUM RBRACKET LPAREN formula_list RPAREN
	  {$$ = new Gate(Gate::tTHRESHOLD, $7);
	   if($2 > $4) cperror("Threshold min > max\n");
	   $$->tmin=$2; $$->tmax=$4; }
	;

%%

static void addGate(char *name, Gate *rightsf) {
  Gate *existing_gate = cp_named_gates[name];
  
  if(rightsf == 0) {
    /* name; definition */
    if(existing_gate) {
      if(existing_gate->type != Gate::tUNDEF)
	cperror("gate '%s' defined twice", name);
      else
	existing_gate->type = Gate::tVAR;
    } else /*existing_gate == 0*/ {
      cp_named_gates[name] = new Gate(Gate::tVAR, name);
    }
    return;
  }

  /* rightsf != 0 */

  if(existing_gate) {
    DEBUG_ASSERT(!existing_gate->names.empty());
    if(existing_gate->type != Gate::tUNDEF)
      cperror("gate '%s' defined twice", name);
    if(!rightsf->names.empty()) {
      if(existing_gate == rightsf) {
	/* 'a := a' definitions */
	rightsf->type = Gate::tVAR;
	free(name);
	return;
      }
      else {
	existing_gate->type = Gate::tREF;
	DEBUG_ASSERT(existing_gate->children == 0);
	existing_gate->add_child(rightsf);
	return;
      }
    }
    else {
      /* rightsf->names.empty() */
      DEBUG_ASSERT(rightsf->names.empty());
      existing_gate->type = rightsf->type;
      while(rightsf->children)
	rightsf->children->change_father(existing_gate);
      existing_gate->tmin = rightsf->tmin;
      existing_gate->tmax = rightsf->tmax;
      delete rightsf;
      return;
    }
    internal_error("Hmmm???");
  }
  
  /*existing_gate == 0*/
  if(!rightsf->names.empty()) {
    Gate *new_gate = new Gate(Gate::tREF, name);
    new_gate->add_child(rightsf);
    cp_named_gates[name] = new_gate;
    return;
  }

  rightsf->names.push_back(name);
  cp_named_gates[name] = rightsf;
  return;
}

